/*
 * clamz - Command-line downloader for the Amazon.com MP3 store
 * Copyright (c) 2008-2010 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>

#include <curl/curl.h>

#include "clamz.h"

struct _clamz_downloader {
  const clamz_config *cfg;
  CURL *curl;
  char *filename;
  int outfd;
  clamz_track *track;
  int last_progress;
};

/* Initialize downloader state */
clamz_downloader *new_downloader(const clamz_config *cfg)
{
  clamz_downloader *dl = malloc(sizeof(clamz_downloader));
  char *cookiejar;

  if (!dl) {
    print_error("Out of memory");
    return NULL;
  }

  if (!cfg->printonly) {
    dl->curl = curl_easy_init();

    if (!dl->curl) {
      print_error("Unable to initialize curl");
      free(dl);
      return NULL;
    }

    curl_easy_setopt(dl->curl, CURLOPT_NOPROGRESS, 0L);
    curl_easy_setopt(dl->curl, CURLOPT_FAILONERROR, 1L);
    curl_easy_setopt(dl->curl, CURLOPT_FOLLOWLOCATION, 1L);
    curl_easy_setopt(dl->curl, CURLOPT_COOKIEFILE, "");
    curl_easy_setopt(dl->curl, CURLOPT_USERAGENT, PACKAGE_STRING);

    cookiejar = get_config_file_name("cookies");
    if (cookiejar) {
      curl_easy_setopt(dl->curl, CURLOPT_COOKIEJAR, cookiejar);
      free(cookiejar);
    }
  }
  else
    dl->curl = NULL;

  dl->cfg = cfg;
  dl->filename = NULL;
  dl->outfd = -1;
  dl->track = NULL;
  return dl;
}

/* Free downloader state */
void free_downloader(clamz_downloader *dl)
{
  if (dl->curl)
    curl_easy_cleanup(dl->curl);
  if (dl->filename)
    free(dl->filename);
  if (dl->outfd > -1)
    close(dl->outfd);
  free(dl);
}

/* Convert a string according to user's preferences */
static char *convert_string(const clamz_config *cfg, const char *s)
{
  const unsigned char *us = (const unsigned char*) s;
  char *converted;
  int len;
  int i, j;

  len = strlen(s) + 1;
  converted = malloc(len * sizeof(char));

  if (!converted) {
    print_error("Out of memory");
    return NULL;
  }

  i = j = 0;
  while (s[i]) {
    if (j >= len - 1) {
      len *= 2;
      converted = realloc(converted, len * sizeof(char));

      if (!converted) {
	print_error("Out of memory");
	return NULL;
      }
    }

    if (us[i] & 0x80) {
      if (cfg->allowutf8) {
	converted[j++] = s[i++];
      }
      else {
	converted[j++] = '_';
	do
	  i++;
	while (((unsigned char) s[i] & 0xc0) == 0x80);
      }
    }
    else if (s[i] == '/' || iscntrl(us[i])) {
      converted[j++] = '_';
      i++;
    }
    else if (isupper(us[i]) && !cfg->allowupper) {
      converted[j++] = tolower(s[i++]);
    }
    else if (cfg->forbid_chars && strchr(cfg->forbid_chars, s[i])) {
      converted[j++] = '_';
      i++;
    }
    else {
      converted[j++] = s[i++];
    }
  }

  converted[j] = 0;

  return converted;
}

static int append_subst_str(char **filename, const char *format,
			    clamz_downloader *dl)
{
  const char *p, *q;
  const char *f = format;
  const char *value;
  char *var, *converted;
  int subst_raw;

  while (*f) {
    if ((p = strchr(f, '$'))) {
      if (p != f
	  && concatenate(filename, f, p - f))
	return 1;

      p++;
      if (*p == '{') {
	p++;
	q = strchr(p, '}');
	if (!q) {
	  print_error("Invalid variable reference in '%s'", format);
	  return 1;
	}
	f = q + 1;
      }
      else {
	q = p;
	while (isalnum((int) (unsigned char) *q) || *q == '_')
	  q++;
	f = q;
      }

      subst_raw = 0;

      if (p == q)
	value = "$";
      else {
	var = NULL;
	if (concatenate(&var, p, q - p))
	  return 1;

	if (!strcasecmp(var, "title"))
	  value = dl->track->title;
	else if (!strcasecmp(var, "creator"))
	  value = dl->track->creator;
	else if (!strcasecmp(var, "album"))
	  value = dl->track->album;
	else if (!strcasecmp(var, "tracknum")) {
	  value = dl->track->trackNum;
	  if (value[0] && !value[1]) {
	    if (concatenate(filename, "0", 1))
	      return 1;
	  }
	}
	else if (!strcasecmp(var, "album_artist"))
	  value = find_meta(dl->track->meta, TMETA_ALBUM_ARTIST);
	else if (!strcasecmp(var, "genre"))
	  value = find_meta(dl->track->meta, TMETA_GENRE);
	else if (!strcasecmp(var, "discnum"))
	  value = find_meta(dl->track->meta, TMETA_DISC_NUM);
	else if (!strcasecmp(var, "suffix")) {
	  value = find_meta(dl->track->meta, TMETA_TRACK_TYPE);
	  if (!value)
	    value = "mp3";
	}
	else if (!strcasecmp(var, "asin"))
	  value = find_meta(dl->track->meta, TMETA_ASIN);
	else if (!strcasecmp(var, "album_asin"))
	  value = find_meta(dl->track->meta, TMETA_ALBUM_ASIN);
	else if (!strcasecmp(var, "amz_title"))
	  value = dl->track->playlist->title;
	else if (!strcasecmp(var, "amz_creator"))
	  value = dl->track->playlist->creator;
	else if (!strcasecmp(var, "amz_asin"))
	  value = find_meta(dl->track->playlist->meta, PMETA_ASIN);
	else if (!strcasecmp(var, "amz_genre"))
	  value = find_meta(dl->track->playlist->meta, PMETA_GENRE);
	else {
	  value = getenv(var);
	  subst_raw = 1;
	}

	free(var);
      }

      if (value) {
	if (subst_raw) {
	  if (concatenate(filename, value, strlen(value)))
	    return 1;
	}
	else {
	  converted = convert_string(dl->cfg, value);
	  if (!converted)
	    return 1;
	  if (concatenate(filename, converted, strlen(converted))) {
	    free(converted);
	    return 1;
	  }
	  free(converted);
	}
      }
    }
    else {
      return concatenate(filename, f, strlen(f));
    }
  }

  return 0;
}

static int create_parents(char *filename)
{
  char *p;

  p = filename;
  while (p && *p) {
    p = strchr(p + 1, '/');
    if (p) {
      *p = 0;
      if (mkdir(filename, 0777) && errno != EEXIST) {
	print_error("Cannot create directory %s: %s", filename,
		    strerror(errno));
	*p = '/';
	return 1;
      }
      *p = '/';
      p++;
    }
  }

  return 0;
}


static size_t write_output(void *ptr, size_t size, size_t n, void *data)
{
  clamz_downloader *dl = data;
  int r;

  r = write(dl->outfd, ptr, size * n);
  if (r < 0) {
    print_error("Error writing to %s: %s", dl->filename, strerror(errno));
    return 0;
  }
  else
    return r;
}


static int show_progress(void *data, double dltotal, double dlnow,
			 double ultotal UNUSED, double ulnow UNUSED)
{
  clamz_downloader *dl = data;
  int progress;

  if (dl->cfg->quiet)
    return 0;

  if (dltotal > 0)
    progress = (int) (100 * dlnow / dltotal);
  else
    progress = -1;

  if (progress != dl->last_progress) {
    dl->last_progress = progress;
    print_progress(dl->track, dl->filename, progress);
  }

  return 0;
}

int download_track(clamz_downloader *dl, clamz_track *tr)
{
  int i;
  char *s;
  CURLcode err;
  off_t startpos;

  if (!tr->location) {
    print_error("No URL provided for this track");
    return 2;
  }

  dl->track = tr;

  if (dl->filename)
    free(dl->filename);
  dl->filename = NULL;

  /* ignore output_dir if name_format is an absolute path */
  if (dl->cfg->output_dir
      && (!dl->cfg->name_format || dl->cfg->name_format[0] != '/')) {
    if (append_subst_str(&dl->filename, dl->cfg->output_dir, dl))
      return 1;
    if (append_subst_str(&dl->filename, "/", dl))
      return 1;
  }

  if (dl->cfg->name_format) {
    if (append_subst_str(&dl->filename, dl->cfg->name_format, dl))
      return 1;
  }
  else {
    print_error("No output filename specified");
    return 1;
  }

  if (!dl->cfg->resume && !access(dl->filename, F_OK)) {
    s = malloc((strlen(dl->filename) + 10) * sizeof(char));
    if (!s) {
      print_error("Out of memory");
      return 1;
    }

    i = 1;
    do {
      sprintf(s, "%s.%d", dl->filename, i);
      i++;
    } while (!access(s, F_OK));

    print_error("\"%s\" already exists; renaming new file to \"%s\"",
		dl->filename, s);

    free(dl->filename);
    dl->filename = s;
  }

  if (dl->cfg->printonly) {
    printf("  Output to \"%s\"\n", dl->filename);
    free(dl->filename);
    dl->filename = NULL;
    return 0;
  }

  if (create_parents(dl->filename)) {
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }
 
  dl->outfd = open(dl->filename, O_WRONLY | O_APPEND | O_CREAT, 0666);

  if (dl->outfd < 0) {
    print_error("Unable to open \"%s\" (%s)", dl->filename, strerror(errno));

    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  /* NOTE: there isn't any way to determine the file size before
     starting the download.
     - The file size declared in the amz file (TMETA_FILE_SIZE) is wrong.
     - Amazon's servers apparently forbid HEAD requests.
  */

  if (!dl->cfg->quiet)
    fprintf(stderr, "Downloading \"%s\"\n", dl->filename);

  i = 0;
  do {
    i++;
    dl->last_progress = -2;

    curl_easy_setopt(dl->curl, CURLOPT_WRITEFUNCTION, write_output);
    curl_easy_setopt(dl->curl, CURLOPT_WRITEDATA, dl);

    curl_easy_setopt(dl->curl, CURLOPT_PROGRESSFUNCTION, show_progress);
    curl_easy_setopt(dl->curl, CURLOPT_PROGRESSDATA, dl);

    curl_easy_setopt(dl->curl, CURLOPT_URL, tr->location);

    startpos = lseek(dl->outfd, (off_t) 0, SEEK_END);
    curl_easy_setopt(dl->curl, CURLOPT_RESUME_FROM_LARGE,
		     (curl_off_t) startpos);

    err = curl_easy_perform(dl->curl);

    if (!err) {
      /* success! */
      break;
    }

    if (startpos != 0 && err == CURLE_HTTP_RANGE_ERROR) {
      /* assume that this means we've already downloaded the whole
	 thing... I guess */
      print_progress(tr, dl->filename, 100);
      err = 0;
      break;
    }

    print_error("Error downloading file: %s", curl_easy_strerror(err));
    if (i < dl->cfg->maxattempts)
      sleep(2);

  } while (i < dl->cfg->maxattempts);

  if (err) {
    close(dl->outfd);
    dl->outfd = -1;
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  if (close(dl->outfd)) {
    print_error("Error writing to %s", dl->filename);
    dl->outfd = -1;
    free(dl->filename);
    dl->filename = NULL;
    return 4;
  }

  dl->outfd = -1;
  free(dl->filename);
  dl->filename = NULL;
  return 0;
}
